#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
      @file: kafka_consumer.py
   @created: Wed, 30 Jun 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import threading
from time import sleep
from typing import List

from PyQt5.QtCore import pyqtSignal, QThread
from confluent_kafka.cimpl import Consumer

from hspylib.core.tools.commons import syserr


# Example at https://docs.confluent.io/platform/current/tutorials/examples/clients/docs/python.html
# For all kafka settings: https://github.com/edenhill/librdkafka/blob/master/CONFIGURATION.md
class KafkaConsumer(QThread):
    """TODO"""

    messageConsumed = pyqtSignal(str, int, int, bytes)

    def __init__(self, poll_interval: float = 0.5):
        super().__init__()
        self.setObjectName('kafka-consumer')
        self._started = False
        self._poll_interval = poll_interval
        self._consumer = None
        self._worker_thread = None
        self.start()

    def start_consumer(self, settings: dict) -> None:
        """Start the Kafka consumer agent"""
        if self._consumer is None:
            self._consumer = Consumer(settings)
            self._started = True

    def stop_consumer(self) -> None:
        """Stop the Kafka consumer agent"""
        if self._consumer is not None:
            self._started = False

    def consume(self, topics: List[str]) -> None:
        """Start the consumer thread"""
        if self._started:
            self._worker_thread = threading.Thread(target=self._consume, args=(topics,))
            self._worker_thread.name = f"kafka-consumer-worker-{hash(self)}"
            self._worker_thread.setDaemon(True)
            self._worker_thread.start()

    def is_started(self):
        """Whether the consumer is started or not"""
        return self._started

    def run(self) -> None:
        while not self.isFinished():
            sleep(self._poll_interval)

    def _consume(self, topics: List[str]) -> None:
        """Consume messages from the selected Kafka topics"""
        try:
            self._consumer.subscribe(topics)
            while self._started and self._consumer is not None:
                message = self._consumer.poll(self._poll_interval)
                if message is None:
                    continue
                elif message.error():
                    syserr(f"An error occurred consuming from Kafka: {message.error().str()}")
                else:
                    self.messageConsumed.emit(
                        message.topic(), message.partition(), message.offset(), message.value()
                    )
        except KeyboardInterrupt:
            syserr("Keyboard interrupted")
        finally:
            if self._consumer:
                self._consumer.close()
                del self._consumer
                self._consumer = None
                self._worker_thread = None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   hspylib.main.hspylib.modules.mock
      @file: mock_request.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""
from typing import Any

from requests.structures import CaseInsensitiveDict

from hspylib.core.enums.charset import Charset
from hspylib.core.enums.content_type import ContentType
from hspylib.core.enums.http_code import HttpCode
from hspylib.core.enums.http_method import HttpMethod
from hspylib.modules.fetch.http_response import HttpResponse


class MockResponse(HttpResponse):
    """TODO"""

    def __init__(self,
                 parent,
                 method: HttpMethod,
                 url: str,
                 status_code: HttpCode = None,
                 body: str = None,
                 headers=None,
                 encoding: Charset = Charset.UTF_8,
                 content_type=ContentType.APPLICATION_JSON):

        super().__init__(method, url, status_code, body, headers, encoding, content_type)
        self.parent = parent
        self.received_body = False

    def then_return(
        self,
        code: HttpCode,
        body: str = None,
        headers = None,
        encoding: Charset = Charset.UTF_8,
        content_type=ContentType.APPLICATION_JSON) -> Any:
        """TODO"""

        response = self.parent.mock(self.method, self.url)
        response.status_code = code
        response.body = body
        response.headers = headers if headers else []
        response.encoding = encoding
        response.content_type = content_type
        if response.content_type:
            response.content_type.charset = encoding

        return self.parent

    def then_return_with_received_body(
        self,
        code: HttpCode,
        headers: CaseInsensitiveDict = None,
        encoding: Charset = Charset.UTF_8,
        content_type=ContentType.APPLICATION_JSON) -> Any:
        """TODO"""

        response = self.parent.mock(self.method, self.url)
        response.received_body = True
        response.body = None
        response.status_code = code
        response.headers = headers if headers else []
        response.encoding = encoding
        response.content_type = content_type
        if response.content_type:
            response.content_type.charset = encoding

        return self.parent



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   hspylib.main.hspylib.modules.qt.views
      @file: qt_view.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""
import os
from abc import ABC
from typing import Tuple, Type, Optional

from PyQt5 import uic
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QWidget

from hspylib.core.tools.commons import run_dir


class QtView(ABC):
    """TODO"""

    @staticmethod
    def load_form(
        form_file: str,
        load_dir: str = f"{run_dir()}/resources/forms/") -> Tuple[Type, Type]:
        """Load the ui form from the .ui file"""

        assert os.path.exists(load_dir) and os.path.isdir(load_dir), \
            f"Load dir {load_dir} does not exist or is not a folder"
        filepath = f"{load_dir}/{form_file}"
        assert os.path.exists(filepath) and os.path.isfile(filepath) and filepath.lower().endswith('.ui'), \
            f"Form file {form_file} does not exist or it not a valid UI form file"

        return uic.loadUiType(filepath)

    def __init__(self, ui_file: str = 'main_qt_view.ui', parent: Optional[QWidget] = None):
        ui_clazz, window_clazz = self.load_form(ui_file)
        # Must come after the initialization above
        self.window, self.ui = window_clazz(), ui_clazz()
        self.ui.setupUi(self.window)
        self.parent = parent

    def show(self) -> None:
        """Show the main widget"""
        self.window.show()

    def set_default_font(self, default_font: QFont = QFont('Courier New', 14)):
        """Set font for all UI components at once"""
        widgets = list(filter(lambda o: hasattr(getattr(self.ui, o), 'setFont'), vars(self.ui)))
        list(map(lambda w: getattr(self.ui, w).setFont(default_font), widgets))



# _*_ coding: utf-8 _*_
#
# HSPyLib v0.11.0
#
# Package: main.hspylib.modules.qt.kafka
"""Package initialization."""

__all__ = [
  'kafka_consumer', 
  'kafka_message', 
  'kafka_producer', 
  'kafka_statistics'
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   hspylib.main.hspylib.modules.mock
      @file: mock_server.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

from http.server import HTTPServer
from random import randint
from threading import Thread
from typing import Any, Optional, Tuple

from hspylib.core.enums.http_method import HttpMethod
from hspylib.modules.mock.mock_request import MockResponse
from hspylib.modules.mock.mock_server_handler import MockServerHandler


class MockServer(HTTPServer):
    """TODO"""

    RANDOM_PORT = randint(49152, 65535)

    def __init__(self, hostname: str, port: int):
        self._mocks = {}
        self.hostname = hostname
        self.port = port
        self.version = '0.9.0'
        super().__init__(self.address(), MockServerHandler)

    def mock(self, method: HttpMethod, url: str) -> Optional[MockResponse]:
        """TODO"""
        try:
            return self._mocks[method][url]
        except KeyError:
            return None

    def address(self) -> Tuple[str, int]:
        """TODO"""
        return self.hostname, self.port

    def is_allowed(self, method: HttpMethod) -> bool:
        """TODO"""
        return method in self._mocks

    def start(self) -> None:
        """TODO"""
        runner = ServerThread(self)
        runner.start()

    def stop(self) -> None:
        """TODO"""
        self.shutdown()
        self.server_close()

    def when_request(self, method: HttpMethod, url: str) -> Any:
        """TODO"""
        request = self.mock(method=method, url=url) or MockResponse(self, method, url)
        self._mocks[method] = self._mocks[method] \
            if method in self._mocks else {}
        self._mocks[method][url] = request
        return request


class ServerThread(Thread):
    def __init__(self, parent: MockServer):
        super().__init__()
        self.parent = parent

    def run(self) -> None:
        self.parent.serve_forever()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
   hspylib.main.hspylib.modules.qt.promotions
      @file: htablemodel.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import collections
import logging as log
from typing import Type, List, TypeVar

from PyQt5.QtCore import QAbstractTableModel, QModelIndex, QVariant, Qt
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QTableView

from hspylib.core.tools.commons import class_attribute_names, class_attribute_values

T = TypeVar('T')

class HTableModel(QAbstractTableModel):
    """TODO"""

    def __init__(
        self,
        parent: QTableView,
        clazz: Type,
        headers: List[str] = None,
        table_data: List[T] = None,
        cell_alignments: List[Qt.AlignmentFlag] = None,
        max_rows: int = 1000):

        QAbstractTableModel.__init__(self, parent)
        self.clazz = clazz
        self.table_data = collections.deque(maxlen=max_rows)
        list(map(self.table_data.append, table_data or []))
        self.headers = headers or self._headers_by_entity()
        self.cell_alignments = cell_alignments
        parent.setModel(self)
        log.info('{} table_headers={}'.format(clazz.__class__.__name__, '|'.join(self.headers)))

    def row(self, index: QModelIndex) -> T:
        """TODO"""
        return self.table_data[index.row()]

    def column(self, index: QModelIndex) -> T:
        """TODO"""
        return class_attribute_values(self.table_data[index.row()].__dict__)[index.column()]

    def push_data(self, data: List[T]):
        """TODO"""
        list(map(self.table_data.append, data))
        self.layoutChanged.emit()

    def clear(self):
        """TODO"""
        self.table_data.clear()
        self.layoutChanged.emit()

    def sort(self, column: int, order: Qt.SortOrder = ...) -> None:
        """TODO"""
        keys = class_attribute_names(self.clazz)
        self.table_data = sorted(self.table_data, key=lambda x: getattr(x, keys[column]), reverse=bool(order))
        self.layoutChanged.emit()

    def data(self, index: QModelIndex, role: int = ...) -> QVariant:
        """TODO"""
        ret_val = QVariant()
        entity = class_attribute_values(self.table_data[index.row()].__dict__)[index.column()]
        str_entity = str(entity) if entity is not None else ''
        if role == Qt.DisplayRole:
            ret_val = \
                QVariant(str_entity)
        if role == Qt.TextAlignmentRole:
            ret_val = \
                QVariant(self.cell_alignments[index.column()]) \
                    if self.cell_alignments else QVariant(Qt.AlignLeft | Qt.AlignVCenter)
        if role == Qt.BackgroundColorRole:
            ret_val = \
                QVariant(QColor(57, 57, 57) if index.row() % 2 == 0 else '')

        return ret_val

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = ...) -> QVariant:
        """TODO"""
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return QVariant(self.headers[section].upper()) if len(self.headers) >= section else QVariant('-')
        if orientation == Qt.Vertical and role == Qt.DisplayRole:
            return QVariant(str(section))
        return QVariant()

    def rowCount(self, parent: QModelIndex = ...) -> int:  # pylint: disable=unused-argument
        """TODO"""
        return len(self.table_data) if self.table_data and len(self.table_data) > 0 else 0

    def columnCount(self, parent: QModelIndex = ...) -> int:  # pylint: disable=unused-argument
        """TODO"""
        return len(self.table_data[0].__dict__.keys()) if self.table_data and len(self.table_data) > 0 else 0

    def _headers_by_entity(self) -> List[str]:
        """TODO"""
        attributes = class_attribute_names(self.clazz)
        return [str(x).capitalize() for x in attributes]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
      @file: htableview.py
   @created: Wed, 30 Jun 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

from typing import Optional

import pyperclip as pyperclip
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QPainter, QPaintEvent, QCursor
from PyQt5.QtWidgets import QTableView, QWidget, QMenu, QHeaderView, QAbstractScrollArea


class HTableView(QTableView):
    """TODO"""

    def __init__(self, parent: Optional[QWidget], placeholder: Optional[str] = None):
        super().__init__(parent)
        self.placeholder = placeholder or 'No data to display'
        self.customContextMenuRequested.connect(self._context_menu)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.horizontalHeader().setStretchLastSection(True)
        self.verticalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.setSizeAdjustPolicy(QAbstractScrollArea.AdjustToContents)

    def clear(self):
        """TODO"""
        self.model().clear()

    def paintEvent(self, event: QPaintEvent) -> None:
        """TODO"""
        super().paintEvent(event)
        if self.model() is not None and self.model().rowCount() > 0:
            return
        color = self.palette().placeholderText().color()
        painter = QPainter(self.viewport())
        painter.save()
        painter.setPen(color)
        elided_text = self.fontMetrics()\
            .elidedText(self.placeholder, Qt.ElideRight, self.viewport().width())
        painter.drawText(self.viewport().rect(), Qt.AlignCenter, elided_text)
        painter.restore()

    def copy(self) -> None:
        """TODO"""
        text = ''

        index_list = self.selectionModel().selectedIndexes()
        last_row = 0
        for index in index_list:
            if len(text) > 0:
                if last_row == index.row():
                    text += ' '
                else:
                    text += '\n'
            text += str(self.model().column(index))
            last_row = index.row()

        pyperclip.copy(text)

    def _context_menu(self):
        """Display the custom context menu"""
        self._menu = QMenu(self)
        self._menu.addAction(u'Copy', self.copy)
        self._menu.addSeparator()
        self._menu.addAction(u'Clear', self.clear)
        self._menu.exec_(QCursor.pos())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
      @file: kafka_statistics.py
   @created: Wed, 30 Jun 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

from typing import Tuple

from PyQt5.QtCore import QThread, pyqtSignal

from hspylib.core.tools.commons import now_ms


class KafkaStatistics(QThread):
    """TODO"""

    statisticsReported = pyqtSignal(int, int, int, int, int, int)

    class ConsumerStats:
        """Gather consumer statistics"""

        def __init__(self):
            self.total = 0
            self.in_a_tick = 0

    class ProducerStats:
        """Gather producer statistics"""

        def __init__(self):
            self.total = 0
            self.in_a_tick = 0

    def __init__(self, report_interval: int = 1):
        super().__init__()
        self.setObjectName('kafka-statistics')
        self._started_ts = now_ms()
        self._consumed = self.ConsumerStats()
        self._produced = self.ProducerStats()
        self._report_interval = report_interval

    def report_consumed(self, amount: int = 1) -> None:
        """Report a consumed message"""
        self._consumed.in_a_tick += amount
        self._consumed.total += amount

    def report_produced(self, amount: int = 1) -> None:
        """Report a produced message"""
        self._produced.in_a_tick += amount
        self._produced.total += amount

    def get_total(self) -> Tuple[int, int]:
        """Retrieve the totals produced/consumed so far"""
        return self._produced.total, self._consumed.total

    def get_in_a_tick(self) -> Tuple[int, int]:
        """Retrieve the amount produced/consumed in a tick"""
        return self._produced.in_a_tick, self._consumed.in_a_tick

    def run(self) -> None:
        while not self.isFinished():
            self.sleep(self._report_interval)
            self._tick()

    def _tick(self) -> None:
        """TODO"""
        diff_time = max(1, int(now_ms() - self._started_ts))
        self._produced.in_a_tick = 0
        self._consumed.in_a_tick = 0
        self.statisticsReported.emit(
            self._produced.total,
            self._consumed.total,
            self._produced.in_a_tick,
            self._consumed.in_a_tick,
            int(self._produced.total / diff_time),
            int(self._consumed.total / diff_time)
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
      @file: kafka_producer.py
   @created: Wed, 30 Jun 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""

import threading
from time import sleep
from typing import List

from PyQt5.QtCore import pyqtSignal, QThread
from confluent_kafka.cimpl import Producer, KafkaError, Message

from hspylib.core.tools.commons import syserr


# Example at: # Example at https://docs.confluent.io/platform/current/tutorials/examples/clients/docs/python.html
# For all kafka settings: https://github.com/edenhill/librdkafka/blob/master/CONFIGURATION.md
class KafkaProducer(QThread):
    """TODO"""

    messageProduced = pyqtSignal(str, int, int, bytes)

    def __init__(self, poll_interval: float = 0.5, flush_timeout: int = 30):
        super().__init__()
        self.setObjectName('kafka-producer')
        self._started = False
        self._poll_interval = poll_interval
        self._flush_timeout = flush_timeout
        self._producer = None
        self._worker_thread = None
        self.start()

    def start_producer(self, settings: dict) -> None:
        """Start the producer"""
        if self._producer is None:
            self._producer = Producer(settings)
            self._started = True

    def stop_producer(self) -> None:
        """Stop the producer."""
        if self._producer is not None:
            self._purge()
            self._flush()
            self._started = False
            del self._producer
            self._producer = None
            self._worker_thread = None

    def produce(self, topics: List[str], messages: List[str]) -> None:
        """Create a worker thread to produce the messages to the specified topics."""
        if self._started and self._producer is not None:
            self._worker_thread = threading.Thread(target=self._produce, args=(topics, messages,))
            self._worker_thread.name = f"kafka-producer-worker-{hash(self)}"
            self._worker_thread.setDaemon(True)
            self._worker_thread.start()

    def is_started(self) -> bool:
        """Whether the producer is started or not."""
        return self._started

    def run(self) -> None:
        while not self.isFinished():
            sleep(self._poll_interval)

    def _flush(self, timeout: int = 0) -> None:
        """Wait for all messages in the Producer queue to be delivered."""
        if self._producer:
            self._producer.flush(timeout=timeout)

    def _purge(self) -> None:
        """Purge messages currently handled by the producer instance."""
        if self._producer:
            self._producer.purge()

    def _produce(self, topics: List[str], messages: List[str]) -> None:
        """Produce message to topic."""
        try:
            for topic in topics:
                for msg in messages:
                    if msg:
                        self._producer.produce(topic, msg, callback=self._cb_message_produced)
                self._producer.poll(self._poll_interval)
        except KeyboardInterrupt:
            syserr("Keyboard interrupted")
        finally:
            self._producer.flush(self._flush_timeout)

    def _cb_message_produced(self, err: KafkaError, message: Message) -> None:
        """Delivery report handler called on successful or failed delivery of message"""
        if err is not None:
            syserr(f"Failed to deliver message: {err}")
        else:
            self.messageProduced.emit(
                message.topic(), message.partition(), message.offset(), message.value()
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   TODO Purpose of the file
   @project: HSPyLib
      @file: kafka_message.py
   @created: Wed, 30 Jun 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2021, HSPyLib team
"""


class KafkaMessage:
    """Represent a produced or consumed Kafka message"""

    def __init__(
        self,
        timestamp: int = None,
        topic: str = None,
        partition: int = None,
        offset: int = None,
        value: str = None):

        self.timestamp = timestamp
        self.topic = topic
        self.partition = partition
        self.offset = offset
        self.value = value

    def __str__(self):
        return f"{self.timestamp} topic={self.topic} partition={self.partition} offset={self.offset} value={self.value}"

    def __repr__(self):
        return str(self)
